//
//  FYMASConstraintMaker.h
//  Masonry
//
//  Created by Jonas Budelmann on 20/07/13.
//  Copyright (c) 2013 cloudling. All rights reserved.
//

#import "FYMASConstraint.h"
#import "FYMASUtilities.h"

typedef NS_OPTIONS(NSInteger, FYMASAttribute) {
    FYMASAttributeLeft = 1 << NSLayoutAttributeLeft,
    FYMASAttributeRight = 1 << NSLayoutAttributeRight,
    FYMASAttributeTop = 1 << NSLayoutAttributeTop,
    FYMASAttributeBottom = 1 << NSLayoutAttributeBottom,
    FYMASAttributeLeading = 1 << NSLayoutAttributeLeading,
    FYMASAttributeTrailing = 1 << NSLayoutAttributeTrailing,
    FYMASAttributeWidth = 1 << NSLayoutAttributeWidth,
    FYMASAttributeHeight = 1 << NSLayoutAttributeHeight,
    FYMASAttributeCenterX = 1 << NSLayoutAttributeCenterX,
    FYMASAttributeCenterY = 1 << NSLayoutAttributeCenterY,
    FYMASAttributeBaseline = 1 << NSLayoutAttributeBaseline,

    FYMASAttributeFirstBaseline = 1 << NSLayoutAttributeFirstBaseline,
    FYMASAttributeLastBaseline = 1 << NSLayoutAttributeLastBaseline,
    
#if TARGET_OS_IPHONE || TARGET_OS_TV
    
    FYMASAttributeLeftMargin = 1 << NSLayoutAttributeLeftMargin,
    FYMASAttributeRightMargin = 1 << NSLayoutAttributeRightMargin,
    FYMASAttributeTopMargin = 1 << NSLayoutAttributeTopMargin,
    FYMASAttributeBottomMargin = 1 << NSLayoutAttributeBottomMargin,
    FYMASAttributeLeadingMargin = 1 << NSLayoutAttributeLeadingMargin,
    FYMASAttributeTrailingMargin = 1 << NSLayoutAttributeTrailingMargin,
    FYMASAttributeCenterXWithinMargins = 1 << NSLayoutAttributeCenterXWithinMargins,
    FYMASAttributeCenterYWithinMargins = 1 << NSLayoutAttributeCenterYWithinMargins,

#endif
    
};

/**
 *  Provides factory methods for creating FYMASConstraints.
 *  Constraints are collected until they are ready to be installed
 *
 */
@interface FYMASConstraintMaker : NSObject

/**
 *	The following properties return a new FYMASViewConstraint
 *  with the first item set to the makers associated view and the appropriate FYMASViewAttribute
 */
@property (nonatomic, strong, readonly) FYMASConstraint *left;
@property (nonatomic, strong, readonly) FYMASConstraint *top;
@property (nonatomic, strong, readonly) FYMASConstraint *right;
@property (nonatomic, strong, readonly) FYMASConstraint *bottom;
@property (nonatomic, strong, readonly) FYMASConstraint *leading;
@property (nonatomic, strong, readonly) FYMASConstraint *trailing;
@property (nonatomic, strong, readonly) FYMASConstraint *width;
@property (nonatomic, strong, readonly) FYMASConstraint *height;
@property (nonatomic, strong, readonly) FYMASConstraint *centerX;
@property (nonatomic, strong, readonly) FYMASConstraint *centerY;
@property (nonatomic, strong, readonly) FYMASConstraint *baseline;

@property (nonatomic, strong, readonly) FYMASConstraint *firstBaseline;
@property (nonatomic, strong, readonly) FYMASConstraint *lastBaseline;

#if TARGET_OS_IPHONE || TARGET_OS_TV

@property (nonatomic, strong, readonly) FYMASConstraint *leftMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *rightMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *topMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *bottomMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *leadingMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *trailingMargin;
@property (nonatomic, strong, readonly) FYMASConstraint *centerXWithinMargins;
@property (nonatomic, strong, readonly) FYMASConstraint *centerYWithinMargins;

#endif

/**
 *  Returns a block which creates a new FYMASCompositeConstraint with the first item set
 *  to the makers associated view and children corresponding to the set bits in the
 *  FYMASAttribute parameter. Combine multiple attributes via binary-or.
 */
@property (nonatomic, strong, readonly) FYMASConstraint *(^attributes)(FYMASAttribute attrs);

/**
 *	Creates a FYMASCompositeConstraint with type FYMASCompositeConstraintTypeEdges
 *  which generates the appropriate FYMASViewConstraint children (top, left, bottom, right)
 *  with the first item set to the makers associated view
 */
@property (nonatomic, strong, readonly) FYMASConstraint *edges;

/**
 *	Creates a FYMASCompositeConstraint with type FYMASCompositeConstraintTypeSize
 *  which generates the appropriate FYMASViewConstraint children (width, height)
 *  with the first item set to the makers associated view
 */
@property (nonatomic, strong, readonly) FYMASConstraint *size;

/**
 *	Creates a FYMASCompositeConstraint with type FYMASCompositeConstraintTypeCenter
 *  which generates the appropriate FYMASViewConstraint children (centerX, centerY)
 *  with the first item set to the makers associated view
 */
@property (nonatomic, strong, readonly) FYMASConstraint *center;

/**
 *  Whether or not to check for an existing constraint instead of adding constraint
 */
@property (nonatomic, assign) BOOL updateExisting;

/**
 *  Whether or not to remove existing constraints prior to installing
 */
@property (nonatomic, assign) BOOL removeExisting;

/**
 *	initialises the maker with a default view
 *
 *	@param	view	any FYMASConstraint are created with this view as the first item
 *
 *	@return	a new FYMASConstraintMaker
 */
- (id)initWithView:(FYMAS_VIEW *)view;

/**
 *	Calls install method on any FYMASConstraints which have been created by this maker
 *
 *	@return	an array of all the installed FYMASConstraints
 */
- (NSArray *)install;

- (FYMASConstraint * (^)(dispatch_block_t))group;

@end
